import copy
import data
import engine
import items
import rpg

import pygame
import random

class TileSource(object):
    
    NOTHING       = 0
    FLOOR         = 1
    H_WALL        = 2
    H_WALL_END_LL = 3
    V_WALL_LEFT   = 4
    H_WALL_END_LR = 5
    H_WALL_END_UL = 6
    V_WALL_RIGHT  = 7
    H_WALL_END_UR = 8
    ORDINARY_WALL = 9
    
    TILECOUNT = 2
    
    def __init__(self,width,height):
        self.width = width
        self.height= height
        self.clear()
        
    def clear(self, index=0):
        self.tiles = [ [index for x in range(self.width)]
                       for y in range(self.height)]
    
    def allTiles(self):
        for rowNum,row in enumerate(self.tiles):
            for colNum,col in enumerate(row):
                yield rowNum,colNum,self.tiles[rowNum][colNum]
        
    def getTiles(self):
        return self.tiles
    
    def isWalkable(self, rowcol):
        row,col = rowcol
        if(row >= 0 and col >= 0 and
           row < len(self.tiles) and col < len(self.tiles[row])):
            return self.tiles[row][col] == self.FLOOR
        return False
    
    tile = property(getTiles)
        
class RoomSource(TileSource):
    def __init__(self,width,height):
        TileSource.__init__(self,width,height)
        for y in range(self.height):
            for x in range(self.width):
                if x == 0 and y == self.height - 1:
                    self.tiles[y][x] = TileSource.H_WALL_END_LL
                elif x == self.width -1 and y == self.height -1:
                    self.tiles[y][x] = TileSource.H_WALL_END_LR
                elif x == 0 and y == 0:
                    self.tiles[y][x] = TileSource.H_WALL_END_UL
                elif x == self.width - 1 and y == 0:
                    self.tiles[y][x] = TileSource.H_WALL_END_UR
                elif x == 0:
                    self.tiles[y][x] = TileSource.V_WALL_LEFT
                elif x == self.width - 1:
                    self.tiles[y][x] = TileSource.V_WALL_RIGHT
                elif y == 0 or y == self.height - 1:
                    self.tiles[y][x] = TileSource.H_WALL
                else:
                    self.tiles[y][x] = TileSource.FLOOR
                    
class RandomSource(TileSource):
    def __init__(self, width,height, roomChance = 0.5, cells=3):
        TileSource.__init__(self,width,height)
        
        self.cells = set()
        self.connections = {}
        
        self.cellWidth = width / cells
        self.cellHeight = height / cells
        
        while(len(self.cells) < 2):
            for row in range(0, cells):
                for col in range(0, cells):
                    if random.random() < roomChance:
                        self.cells.add( (row, col, self.makeRoom(row, col)))
                        
        listcells = list(self.cells)
        while(not self.isFullyConnected()):
            r1 = r2 = None
            while(r1 == r2):
                r1 = random.choice(listcells)
                r2 = random.choice(listcells)
                if not (r2 in self.connections.setdefault(r1, [])):
                    self.connect(r1, r2)
        self.prettify()
                    
    def connect(self, r1, r2):
        self.connections[r1].append(r2)
        self.connections.setdefault(r2,[]).append(r1)
        
        row, col, dimensions = r1
        rect1 = pygame.Rect(dimensions)
        row,col,dimensions = r2
        rect2 = pygame.Rect(dimensions)
        
        sign = data.sign(rect2.x - rect1.x)
        
        y = rect1.center[1]
        if sign:
            for x in range(rect1.center[0], rect2.center[0],sign):
                self.tiles[y][x] = TileSource.FLOOR
            
        sign = data.sign(rect2.y - rect1.y)
        if sign:
            x = rect2.center[0]
            for y in range(rect1.center[1], rect2.center[1],sign):
                self.tiles[y][x] = TileSource.FLOOR

    def fullyExplore(self, cell, prev=None):
        if prev == None: prev = set()
        result = set( (cell,))
        prev.add(cell)
        row,col,dimensions = cell
        for connection in self.connections.setdefault(cell, []):
            if connection in prev: continue
            explorationResult = self.fullyExplore(connection, prev)
            result.update(explorationResult)
        return result

    def isFullyConnected(self):
        initialCell = random.choice(list(self.cells))
        covered = self.fullyExplore(initialCell)
        return not (covered ^ self.cells)

    def makeRoom(self, row, col):
        x,y = col * self.cellWidth, row * self.cellHeight
        roomX = random.randint(1, self.cellWidth / 2)
        roomY = random.randint(1, self.cellHeight / 2)
        
        maxW = self.cellWidth - roomX
        maxH = self.cellHeight - roomY
        roomW = random.randint(3, maxW)
        roomH = random.randint(3, maxH)
        
        roomX += x
        roomY += y
        result = (roomX,roomY,roomW,roomH)
        
        for y in range(roomY, roomY + roomH):
            for x in range(roomX, roomX + roomW):
                self.tiles[y][x] = TileSource.FLOOR
                
        return result
    
    def neighbors(self, row, col):
        heightInTiles = len(self.tiles)
        widthInTiles = len(self.tiles[0])
        xOffsets = ( -1, 0, 1)
        yOffsets = ( -1, 0, 1)
        for yOffset in yOffsets:
            for xOffset in xOffsets:
                x = col + xOffset
                y = row + yOffset
                if x == y == 0: continue
                if (x >= 0 and y >= 0 and 
                    x < widthInTiles and
                    y < heightInTiles):
                    yield y,x,self.tiles[y][x]
    
    def prettify(self):
        for row, col, tile in self.allTiles():
            if tile: continue
            # Look around
            for r,c,tile in self.neighbors(row,col):
                if tile == TileSource.FLOOR:
                    self.tiles[row][col] = TileSource.ORDINARY_WALL
                    break
        
class TileMap(engine.CompositeActor):
    def __init__(self, source, level,screenSize, log):
        engine.CompositeActor.__init__(self)
        self.center = None
        self.source = source
        self.level = level
        self.screenSize = screenSize
        self.files = [ 'floor.png', 'hWall.png', 'hWall_end_lowerleft.png',
                      'vWall_left.png', 'hWall_end_lowerright.png',
                      'hWall_end_upperleft.png', 'vWall_right.png',
                      'hWall_end_upperright.png', 'wall.png'
                       ]
        self.tileActors = {}
        self.log = log
        self.goodieBag = items.RandomTreasureGenerator()
        
        self.map = []
        self.map.append(None)
        self.scrollX = 0
        self.scrollY = 0
        for filename in self.files:
            self.map.append(data.load_image(filename))
            
        # For simplicity, tile width/height is calibrated
        # from the first tile.
        self.tileWidth = self.map[1].get_width()
        self.tileHeight = self.map[1].get_height()
    
        self.stairs = None
        self.loot = []
        
        for i in range(self.level):
            self.addRandomLoot()
    
    def __getstate__(self):
        odict = self.__dict__.copy()
        del odict['map']
        return odict
    
    def __setstate__(self, dict):
        map = []
        map.append(None)
        for filename in dict['files']:
            map.append(data.load_image(filename))
        dict['map'] = map
        
        self.__dict__.update(dict)
    
    def addLoot(self, loot):
        self.loot.append(loot)
        self.add(loot)
        
    def addRandomLoot(self):
        
        level = max(1, self.level/2)
        loot = self.goodieBag.generateRandomTreasure(level)
        if loot:
            loot = rpg.Loot(self, loot)
            self.addLoot(loot)
            tLoc = self.randomOpenTile()
            wLoc = self.tileToWorld(tLoc)
            loot.world.center = wLoc
    
    def addStairs(self, stairs):
        self.stairs = stairs
        tLoc = self.randomOpenTile()
        stairs.world.center = self.tileToWorld(tLoc)
        self._stairLoc = tLoc
        self.add(stairs)
    
    def getCenterInWorld(self):
        "Returns the actual center of the tilemap in world co-ords."
        return (self.source.width * self.tileWidth  / 2,
                self.source.height* self.tileHeight / 2)
    
    def isOccupied(self, rowcol, exceptions = []):
        row,col = rowcol
        for actor in self.children:
            if actor in exceptions or not actor.blocking: continue
            tRow, tCol = self.worldToTile(actor.world.midbottom)
            if tCol == col and tRow == row:
                return True
        return False
    
    def occupiedBy(self, rowcol, exceptions = []):
        row,col = rowcol
        items = []
        for actor in self.children:
            if((actor in exceptions) or (not actor.blocking)): continue
            tRow,tCol = self.worldToTile(actor.world.midbottom)
            if tCol == col and tRow == row:
                items.append(actor)
        return items
    
    def paint(self,screen):
        # TODO: Some optimization might be nice here
        for row,col,tile in self.source.allTiles():
            img = self.map[tile]
            if(img):
                x = self.scrollX + col* self.tileWidth
                y = (self.scrollY + 
                     row*self.tileHeight - (img.get_height() - self.tileHeight))
                screen.blit(img, (x, y))
            
            if row in self.tileActors:
                # Horribly inefficient, as will paint all actors on this row
                # repeatedly for each tile in the row.
                # TODO: Find a better way if this slows us down.
                actors = self.tileActors[ row ]
                for actor in actors:
                    actor.paint(screen)
                    
    def randomOpenTile(self):
        open = []
        for row,col,tile in self.source.allTiles():
            if tile == TileSource.FLOOR and not self.isOccupied( (row,col) ):
                open.append( (row,col) )
        return random.choice(open)
                    
    def screenToWorld(self, pos):
        x,y = pos
        return (x - self.scrollX, y - self.scrollY)
    
    def stairLoc(self):
        return self._stairLoc
    
    def takeLoot(self, loot):
        self.loot.remove(loot)
        self.remove(loot)
    
    def tileToScreen(self, tLoc):
        wLoc = self.tileToWorld(tLoc)
        return self.worldToScreen(wLoc)
        
    def tileToWorld(self, tLoc):
        tY, tX = tLoc
        return ( tX * self.tileWidth + self.tileWidth / 2,
                 tY * self.tileHeight + self.tileHeight / 2)
                
    def update(self, delay):
        self.tileActors = {}
        deadlist = []
        for actor in self.children:
            actor.update(delay)
            aLoc = actor.world.midbottom
            tLoc = self.worldToTile( aLoc )
            list = self.tileActors.setdefault(tLoc[0],[])
            list.append(actor)
            
            if not actor.translateable: continue
            
            if actor == self.center:
                self.scrollX = self.screenSize[0] / 2 - actor.world.centerx
                self.scrollY = self.screenSize[1] / 2 - actor.world.centery
            
            actor.rect.centerx = self.scrollX + actor.world.centerx
            actor.rect.centery = self.scrollY + actor.world.centery
            if actor.isDead(): deadlist.append(actor)
        
        for actor in deadlist:
            self.children.remove(actor)
            if actor == self.center:
                continue
            # Drop some treasure
            loot = self.goodieBag.generateRandomTreasure(self.center, actor)
            if loot:
                loot = rpg.Loot(self, loot)
                self.addLoot(loot)
                loot.world.center = actor.world.center
            if self.center.target == actor:
                self.center.target = None
            
    def worldToScreen(self, wLoc):
        x,y = wLoc
        return (x + self.scrollX, y + self.scrollY)
            
    def worldToTile(self, wLoc):
        wX,wY = wLoc[0],wLoc[1]
        return ( wY / self.tileHeight, wX / self.tileWidth )
                
    centerInWorld = property(getCenterInWorld) 